@@ -8,36 +8,30 @@ from csv import reader
 from functools import partial
 from io import StringIO
 from itertools import product
+import logging
+from logging import Logger
+from operator import is_not
 import os
 from pathlib import Path
-from typing import (
-    Any,
-    Callable,
-    Coroutine,
-    DefaultDict,
-    Dict,
-    Iterable,
-    List,
-    Optional,
-    Set,
-    Tuple,
-    Union,
-)
-from warnings import warn
+from typing import Any, Coroutine, DefaultDict, Dict, Iterable, List, Optional, Set, Tuple, Union
 from zipfile import BadZipFile
 
+import geopandas as gpd
+from geopandas.geodataframe import GeoDataFrame
 import numpy as np
 import pandas as pd
 from pandas import DataFrame
 from pkg_resources import resource_string
+from typing_extensions import Literal
 from yarl import URL
 
 from .api import CensusAPI, Table, look_up_census_api_key
 from .geography import (
     Geo,
     coerce_polygon_to_multipolygon,
+    convert_geo_id_to_14_chars,
     flatten_geometry,
-    get_geo_codes,
+    get_geo_mappings,
     identify_affgeoid_field,
     load_geodataframe,
 )
@@ -50,14 +44,16 @@ from .utilities import (
     chunk_variables,
     load_annotations_dataframe,
     parse_table_name_from_variable,
-    tidy_variable_label,
-    titleize_text,
     wrap_scalar_value_in_list,
 )
 
+# Initialize logger
+logger: Logger = logging.getLogger(__name__)
+
 # Types
 Tables = List[Table]
 Variables = Dict[Tuple[int, str], dict]
+GazetteerFiles = List[Optional[DataFrame]]
 Shapefiles = List[Optional[Path]]
 
 
@@ -65,7 +61,7 @@ class Query:
     """A query for American Community Survey data from the Census API.
 
     A Query instance can be used to fetch ACS variables, tables, and
-    shapefiles for a given ACS estimate (1-, 3-, or 5-year), year(s),
+    geometry for a given ACS estimate (1-, 3-, or 5-year), year(s), and
     and geographic units.
     """
 
@@ -76,10 +72,8 @@ class Query:
         variables: Union[Iterable, str],
         for_geo: Union[Iterable, str],
         in_geo: Iterable = None,
-        join_geography: bool = True,
-        table: Any = None,  # Deprecated (no longer necessary)
+        geometry: Optional[Literal['points', 'polygons']] = None,
         census_api_key: str = None,
-        verify_ssl: bool = True,
     ):
         if estimate in [1, 3, 5]:
             self.estimate: int = estimate
@@ -88,13 +82,13 @@ class Query:
         self._years: Iterable = wrap_scalar_value_in_list(years)
         self._variables: Iterable = wrap_scalar_value_in_list(variables)
         self.for_geo: Iterable = [Geo(geo) for geo in wrap_scalar_value_in_list(for_geo)]
-        self.in_geo: Iterable = [] if in_geo is None else [
-            Geo(geo) for geo in wrap_scalar_value_in_list(in_geo)
-        ]
-        if join_geography in [True, False]:
-            self.join_geography: bool = join_geography
+        self.in_geo: Iterable = (
+            [] if in_geo is None else [Geo(geo) for geo in wrap_scalar_value_in_list(in_geo)]
+        )
+        if geometry in ['points', 'polygons', None]:
+            self.geometry: Optional[Literal['points', 'polygons']] = geometry
         else:
-            raise ValueError('Please specify a True/False value for join_geography')
+            raise ValueError('Please specify a valid geometry value: points, polygons')
         self.census_api_key: str = look_up_census_api_key(census_api_key)
 
         # Validate query parameters to avoid common pitfalls
@@ -103,8 +97,8 @@ class Query:
         # Initialize invalid variables defaultdict
         self._invalid_variables: DefaultDict[int, list] = defaultdict(list)
 
-        # Create cache directory, plus any enclosing directories, as needed
-        CACHE_DIRECTORY_PATH.mkdir(parents=True, exist_ok=True)
+        # Create cache directory if it doesn't exist
+        CACHE_DIRECTORY_PATH.mkdir(exist_ok=True)
 
     def __repr__(self) -> str:
         attributes = ['estimate', 'years', 'variables', 'for_geo', 'in_geo']
@@ -177,6 +171,8 @@ class Query:
 
                 nest_asyncio.apply()
                 results: Future = asyncio.run(gathered_calls)  # type: ignore
+            else:
+                raise error
 
         # Compile invalid variables
         variables = {}
@@ -185,14 +181,8 @@ class Query:
             if not variable_json.get('label', False):
                 invalid_variable = variable_json['name']
                 self._invalid_variables[year].append(invalid_variable)
-                if year == 2009:
-                    message = (
-                        f'{invalid_variable} is not a recognized variable for {year}. Note that '
-                        'the Census API does not contain 1-year estimate data for 2009'
-                    )
-                else:
-                    message = f'{invalid_variable} is not a recognized variable for {year}'
-                warn(message)
+                message = f'Warning: {invalid_variable} is not a recognized variable for {year}'
+                logger.warning(message)
             else:
                 variables[year, variable_json['name']] = variable_json
         return variables
@@ -214,21 +204,39 @@ class Query:
         tables = list(results)
         return tables
 
-    def get_geography(self) -> Shapefiles:
-        """Get shapefiles for the supplied variables and geography."""
+    def get_gazetteer_files(self) -> GazetteerFiles:
+        """Get Gazetteer files for the supplied years and geographies."""
         # Assemble API calls for concurrent execution
         calls = []
-        years_with_geography = [year for year in self.years if year >= 2013]
+        years_with_gazetteer_files = [year for year in self.years if year >= 2012]
         # Handle multiple for_geo values by year
-        for year, for_geo in product(years_with_geography, self.for_geo):
-            call: Coroutine[Any, Any, Path] = self._census_api.fetch_geography(
-                year, for_geo, self.in_geo
+        for year, for_geo in product(years_with_gazetteer_files, self.for_geo):
+            call: Coroutine[Any, Any, Optional[DataFrame]] = self._census_api.fetch_gazetteer_file(
+                year, for_geo
             )
             calls.append(call)
         # Make concurrent API calls
         results: Future = asyncio.run(self._census_api.gather_calls(calls))
-        geography = list(results)
-        return geography
+        gazetteer_files = list(results)
+        return gazetteer_files
+
+    def get_shapefiles(self) -> Shapefiles:
+        """Get shapefiles for the supplied years and geographies."""
+        # Assemble API calls for concurrent execution
+        calls = []
+        years_with_shapefiles = [year for year in self.years if year >= 2013]
+
+        # Handle multiple for_geo values by year
+        for year, for_geo in product(years_with_shapefiles, self.for_geo):
+            call: Coroutine[Any, Any, Optional[Path]] = self._census_api.fetch_shapefile(
+                year, for_geo, self.in_geo
+            )
+            calls.append(call)
+
+        # Make concurrent API calls
+        results: Future = asyncio.run(self._census_api.gather_calls(calls))
+        shapefiles = list(results)
+        return shapefiles
 
     def convert_variables_to_dataframe(self, variables: Variables) -> DataFrame:
         """Convert Census API variable data to a dataframe."""
@@ -241,34 +249,67 @@ class Query:
         # Drop/rename columns, clean up label/concept values
         extra_columns = {'attributes', 'group', 'limit', 'predicateType'} & set(dataframe.columns)
         dataframe = dataframe.drop(columns=extra_columns).rename(columns={'name': 'variable'})
-        dataframe['label'] = dataframe['label'].map(tidy_variable_label)
-        dataframe['concept'] = dataframe['concept'].fillna('').map(titleize_text)
 
         return dataframe
 
     def convert_tables_to_dataframe(self, tables: Tables) -> DataFrame:
         """Reshape and convert ACS data tables to a dataframe."""
-        geography_types: Iterable[str] = get_geo_codes().keys()
+        geography_types: Iterable[str] = get_geo_mappings('geo_codes').keys()
 
         # Melt each subset to adopt common schema
         subsets = []
         for header, *rows in tables:
             subset = DataFrame(rows, columns=header)
             # Consolidate geography type in a single column
-            geography_columns: Set[str] = (set(geography_types) & set(subset.columns))
+            geography_columns: Set[str] = set(geography_types) & set(subset.columns)
             id_vars = ['NAME', 'GEO_ID', 'geo_type', *geography_columns, 'year']
             melted: DataFrame = subset.melt(id_vars=id_vars).drop(columns=geography_columns)
             subsets.append(melted)
 
-        # Ensure correct computation of percent change and difference
-        dataframe: DataFrame = pd.concat(subsets).sort_values(
-            by=['geo_type', 'variable', 'NAME', 'year']
-        ).reset_index(drop=True)
+        # Ensure correct sort order and value dtype
+        dataframe: DataFrame = (
+            pd.concat(subsets)
+            .sort_values(by=['geo_type', 'variable', 'NAME', 'year'])
+            .reset_index(drop=True)
+        )
         dataframe['value'] = dataframe['value'].astype(float)
 
         return dataframe
 
-    def convert_geography_to_dataframe(self, geography: Shapefiles) -> DataFrame:
+    def convert_gazetteer_files_to_dataframe(
+        self, gazetteer_files: GazetteerFiles
+    ) -> Optional[DataFrame]:
+        """Convert one or more Gazetteer files to a dataframe.
+
+        Skips over null values produced by invalid responses from the
+        Gazetteer file endpoint.
+        """
+        subsets = []
+        gazetteer_tables: Iterable[DataFrame] = filter(partial(is_not, None), gazetteer_files)
+        nad_83_epsg = 4269
+        for gazetteer_table in gazetteer_tables:
+            subset = GeoDataFrame(
+                gazetteer_table,
+                geometry=gpd.points_from_xy(
+                    gazetteer_table['INTPTLONG'], gazetteer_table['INTPTLAT']
+                ),
+            )
+            subset.crs = f'EPSG:{nad_83_epsg}'
+            subset['gazetteer_geo_id'] = subset.apply(
+                lambda row: convert_geo_id_to_14_chars(row['GEOID'], row['gazetteer_geo_type']),
+                axis=1,
+            )
+            subsets.append(subset)
+
+        # Return null value if no dataframes were obtained
+        if not subsets:
+            return None
+
+        # Concatenate dataframes and return
+        dataframe: GeoDataFrame = pd.concat(subsets)
+        return dataframe
+
+    def convert_shapefiles_to_dataframe(self, shapefiles: Shapefiles) -> DataFrame:
         """Convert one or more shapefiles to a dataframe.
 
         Skips over null filepaths produced by invalid responses from the
@@ -281,35 +322,29 @@ class Query:
         os.environ['CPL_ZIP_ENCODING'] = 'UTF-8'
         subsets = []
         # Drop null values (e.g., for not-yet-released shapefiles) from list of filepaths
-        filepaths: Iterable[Path] = filter(None, geography)
+        filepaths: Iterable[Path] = filter(None, shapefiles)
         for filepath in filepaths:
             try:
                 subset = load_geodataframe(filepath)
             except BadZipFile:
-                warn(
-                    f'Failed to load zip file {filepath}. It may be corrupted. You might try '
-                    'clearing your autocensus cache by calling autocensus.clear_cache() or '
+                logger.warning(
+                    f'Warning: Failed to load zip file {filepath}. It may be corrupted. You might '
+                    'try clearing your autocensus cache by calling autocensus.clear_cache() or '
                     f'manually deleting the cache folder at {CACHE_DIRECTORY_PATH}. Continuing…'
                 )
                 continue
             subsets.append(subset)
         dataframe: DataFrame = pd.concat(subsets, ignore_index=True, sort=True)
 
-        # Reproject dataframe from NAD 83 to WGS 84
-        wgs_84_epsg = 4326
-        dataframe['geometry'] = dataframe['geometry'].to_crs(epsg=wgs_84_epsg)
-        dataframe.crs = f'EPSG:{wgs_84_epsg}'
-
         # Geometry columns
-        dataframe['centroid'] = dataframe.centroid
-        dataframe['internal_point'] = dataframe['geometry'].representative_point()
-        dataframe['geometry'] = (
-            dataframe['geometry'].map(coerce_polygon_to_multipolygon).map(flatten_geometry)
-        )
+        if self.geometry == 'polygons':
+            dataframe['geometry'] = (
+                dataframe['geometry'].map(coerce_polygon_to_multipolygon).map(flatten_geometry)
+            )
 
         # Clean up
         affgeoid_field = identify_affgeoid_field(dataframe.columns)
-        columns_to_keep = [affgeoid_field, 'year', 'centroid', 'internal_point', 'geometry']
+        columns_to_keep = [affgeoid_field, 'year', 'geometry']
         dataframe = dataframe.loc[:, columns_to_keep]
         return dataframe
 
@@ -320,36 +355,24 @@ class Query:
         reorders columns.
         """
         # Drop duplicates (some geospatial datasets, like ZCTAs, include redundant rows)
-        geo_names = {'centroid', 'internal_point', 'geometry'}
+        geo_names = {'geometry'}
         non_geo_names: set = set(dataframe.columns) - geo_names
         dataframe = dataframe.drop_duplicates(subset=non_geo_names, ignore_index=True)
 
         # Insert NAs for annotated rows to avoid outlier values like -999,999,999
         dataframe.loc[dataframe['annotation'].notnull(), 'value'] = np.NaN
 
-        # Compute percent change and difference from prior year (if prior year is NA, uses value
-        # from most recent available year)
-        dataframe['percent_change'] = (
-            dataframe.loc[dataframe['value'].notnull()]
-            .groupby(['GEO_ID', 'variable'])['value']
-            .pct_change()
-        )
-        dataframe['difference'] = (
-            dataframe.loc[dataframe['value'].notnull()]
-            .groupby(['GEO_ID', 'variable'])['value']
-            .diff()
-        )
-
         # Create year date column
-        convert_datetime: Callable = partial(pd.to_datetime, format='%Y-%m-%d')
-        dataframe['date'] = dataframe['year'].map('{}-12-31'.format).map(convert_datetime)
+        dataframe['date'] = pd.to_datetime(
+            dataframe['year'].astype('string') + '-12-31', format='%Y-%m-%d'
+        )
 
         # Rename and reorder columns
         names_csv: bytes = resource_string(__name__, 'resources/names.csv')
         csv_reader = reader(StringIO(names_csv.decode('utf-8')))
         next(csv_reader)  # Skip header row
         names: dict = dict(csv_reader)  # type: ignore
-        if self.join_geography is True:
+        if self.geometry in ['points', 'polygons'] and (set(dataframe.columns) & geo_names):
             name_order = [*names.values(), *geo_names]
         else:
             name_order = list(names.values())
@@ -358,7 +381,11 @@ class Query:
         return dataframe
 
     def assemble_dataframe(
-        self, variables: Variables, tables: Tables, geography: Shapefiles
+        self,
+        variables: Variables,
+        tables: Tables,
+        gazetteer_files: GazetteerFiles,
+        shapefiles: Shapefiles,
     ) -> DataFrame:
         """Merge and finalize the query dataframe.
 
@@ -366,30 +393,38 @@ class Query:
         cleans up and finalizes the combined dataframe.
         """
         # Merge tables with variables, annotations
-        print('Merging ACS tables and variables...')
+        logger.info('Merging ACS tables and variables...')
         tables_dataframe: DataFrame = self.convert_tables_to_dataframe(tables)
         variables_dataframe: DataFrame = self.convert_variables_to_dataframe(variables)
         dataframe = tables_dataframe.merge(
             right=variables_dataframe, how='left', on=['variable', 'year']
         )
-        print('Merging annotations...')
+        logger.info('Merging annotations...')
         annotations_dataframe: DataFrame = load_annotations_dataframe()
         dataframe = dataframe.merge(right=annotations_dataframe, how='left', on=['value'])
 
         # Merge geospatial data if included
-        if self.join_geography is True:
-            print('Merging shapefiles...')
-            geography_dataframe: DataFrame = self.convert_geography_to_dataframe(geography)
-            affgeoid_field = identify_affgeoid_field(geography_dataframe.columns)
-            dataframe = dataframe.merge(
-                right=geography_dataframe,
-                how='left',
-                left_on=['GEO_ID', 'year'],
-                right_on=[affgeoid_field, 'year'],
-            )
+        geometry_dataframe: Optional[DataFrame]
+        right_geo_id_field: str
+        if self.geometry in ['points', 'polygons']:
+            if self.geometry == 'points':
+                logger.info('Merging Gazetteer files...')
+                geometry_dataframe = self.convert_gazetteer_files_to_dataframe(gazetteer_files)
+                right_geo_id_field = 'gazetteer_geo_id'
+            else:
+                logger.info('Merging shapefiles...')
+                geometry_dataframe = self.convert_shapefiles_to_dataframe(shapefiles)
+                right_geo_id_field = identify_affgeoid_field(geometry_dataframe.columns)
+            if geometry_dataframe is not None:
+                dataframe = dataframe.merge(
+                    right=geometry_dataframe,
+                    how='left',
+                    left_on=['GEO_ID', 'year'],
+                    right_on=[right_geo_id_field, 'year'],
+                )
 
         # Finalize dataframe
-        print('Finalizing data...')
+        logger.info('Finalizing data...')
         dataframe = self.finalize_dataframe(dataframe)
 
         return dataframe
@@ -401,16 +436,21 @@ class Query:
         a few seconds to several minutes.
         """
         with self.create_census_api_session():
-            print('Retrieving variables...')
+            logger.info('Retrieving variables...')
             variables: Variables = self.get_variables()
-            print('Retrieving ACS tables...')
+            logger.info('Retrieving ACS tables...')
             tables: Tables = self.get_tables()
-            if self.join_geography is True:
-                print('Retrieving shapefiles...')
-                geography: Shapefiles = self.get_geography()
-            else:
-                geography = None  # type: ignore
-        dataframe = self.assemble_dataframe(variables, tables, geography)
+
+            # Add geometry
+            gazetteer_files: GazetteerFiles = []
+            shapefiles: Shapefiles = []
+            if self.geometry == 'points':
+                logger.info('Retrieving Gazetteer files...')
+                gazetteer_files.extend(self.get_gazetteer_files())
+            elif self.geometry == 'polygons':
+                logger.info('Retrieving shapefiles...')
+                shapefiles.extend(self.get_shapefiles())
+        dataframe = self.assemble_dataframe(variables, tables, gazetteer_files, shapefiles)
         return dataframe
 
     def to_socrata(
